@@ -1,10 +1,11 @@
 # Edit Storage Settings of a Guest
 use strict;
 use warnings;
-our (%gui, %signal, %hostspec, %bool, %EMedVar, %vmc);
+our (%gui, %signal, %vmc);
 
-sub setup_edit_dialog_storage() {
-    &addrow_log("Retrieving storage settings for $vmc{Name}...");
+sub setup_edit_dialog_storage {
+    my $vhost = &vhost();
+    &addrow_log("Fetching storage settings for $vmc{Name}...");
     &busy_pointer($gui{dialogEdit}, 1);
     &fill_list_editstorage($vmc{IMachine});
     $gui{menuAttachAdd} = Gtk2::Menu->new();
@@ -12,7 +13,7 @@ sub setup_edit_dialog_storage() {
     $gui{menuAttachDVD} = Gtk2::Menu->new();
     $gui{menuAttachFloppy} = Gtk2::Menu->new();
     $gui{menuitemAttachHD} = Gtk2::MenuItem->new_with_label('Hard Disks');
-    $gui{menuitemAttachDVD} = Gtk2::MenuItem->new_with_label('CDs/DVDs');
+    $gui{menuitemAttachDVD} = Gtk2::MenuItem->new_with_label('Optical Discs');
     $gui{menuitemAttachFloppy} = Gtk2::MenuItem->new_with_label('Floppy Disks');
     $gui{menuitemAttachHD}->set_submenu($gui{menuAttachHD});
     $gui{menuitemAttachDVD}->set_submenu($gui{menuAttachDVD});
@@ -52,15 +53,15 @@ sub setup_edit_dialog_storage() {
     my $IMediumDVDRef = &get_all_media('DVD');
     my $IMediumFloppyRef = &get_all_media('Floppy');
 
-    foreach (sort { lc($a) cmp lc($b) } (keys %$IMediumHDRef)) {
-        my $item = Gtk2::MenuItem->new_with_label($_);
+    foreach (sort { lc($$IMediumHDRef{$a}) cmp lc($$IMediumHDRef{$b}) } (keys %$IMediumHDRef)) {
+        my $item = Gtk2::MenuItem->new_with_label($$IMediumHDRef{$_});
         $gui{menuAttachHD}->append($item);
         $item->show();
-        $item->signal_connect(activate => \&storage_attach_hd, $$IMediumHDRef{$_});
+        $item->signal_connect(activate => \&storage_attach_hd, $_);
     }
 
-    if ($hostspec{dvd}) {
-        foreach my $pdvd (@{$hostspec{dvd}}) {
+    if ($$vhost{dvd}) {
+        foreach my $pdvd (@{$$vhost{dvd}}) {
             my $item = Gtk2::MenuItem->new_with_label('<Server Drive> ' . IMedium_getLocation($pdvd));
             $gui{menuAttachDVD}->append($item);
             $item->show();
@@ -72,19 +73,19 @@ sub setup_edit_dialog_storage() {
         $pdvdsep->show();
     }
 
-    foreach (sort { lc($a) cmp lc($b) } (keys %$IMediumDVDRef)) {
-        my $item = Gtk2::MenuItem->new_with_label($_);
+    foreach (sort { lc($$IMediumDVDRef{$a}) cmp lc($$IMediumDVDRef{$b}) } (keys %$IMediumDVDRef)) {
+        my $item = Gtk2::MenuItem->new_with_label($$IMediumDVDRef{$_});
         $gui{menuAttachDVD}->append($item);
         $item->show();
-        $item->signal_connect(activate => \&storage_attach_dvd, $$IMediumDVDRef{$_});
+        $item->signal_connect(activate => \&storage_attach_dvd, $_);
     }
 
-    if ($hostspec{floppy}) {
-        foreach my $pfloppy (@{$hostspec{floppy}}) {
+    if ($$vhost{floppy}) {
+        foreach my $pfloppy (@{$$vhost{floppy}}) {
             my $item = Gtk2::MenuItem->new_with_label('<Server Drive> ' . IMedium_getLocation($pfloppy));
             $gui{menuAttachFloppy}->append($item);
             $item->show();
-            $item->signal_connect(activate => \&storage_attach_dvd, $pfloppy);
+            $item->signal_connect(activate => \&storage_attach_floppy, $pfloppy);
         }
 
         my $pfloppysep = Gtk2::SeparatorMenuItem->new();
@@ -92,55 +93,64 @@ sub setup_edit_dialog_storage() {
         $pfloppysep->show();
     }
 
-    foreach (sort { lc($a) cmp lc($b) } (keys %$IMediumFloppyRef)) {
-        my $item = Gtk2::MenuItem->new_with_label($_);
+    foreach (sort { lc($$IMediumFloppyRef{$a}) cmp lc($$IMediumFloppyRef{$b}) } (keys %$IMediumFloppyRef)) {
+        my $item = Gtk2::MenuItem->new_with_label($$IMediumFloppyRef{$_});
         $gui{menuAttachFloppy}->append($item);
         $item->show();
-        $item->signal_connect(activate => \&storage_attach_floppy, $$IMediumFloppyRef{$_});
+        $item->signal_connect(activate => \&storage_attach_floppy, $_);
     }
 
     &busy_pointer($gui{dialogEdit}, 0);
-    &addrow_log("Storage settings for $vmc{Name} retrieved.");
+    &addrow_log('Storage settings complete.');
 }
 
 # Displays the create a new hard disk dialog. Creates & attaches based on options
-sub show_dialog_createhd() {
+sub show_dialog_createhd {
+    my $vhost = &vhost();
     my $guestname = IMachine_getName($vmc{IMachine});
     $gui{comboboxCreateHDFormat}->set_active(0);
     $gui{radiobuttonCreateHDDynamic}->show();
     $gui{radiobuttonCreateHDFixed}->show();
     $gui{radiobuttonCreateHDSplit}->hide();
     $gui{entryCreateHDName}->set_text($guestname . int(rand(9999)));
-    $gui{spinbuttonCreateHDSize}->set_range(8.00, int($hostspec{maxhdsize}/1048576));
+    $gui{spinbuttonCreateHDSize}->set_range(8.00, int($$vhost{maxhdsize}/1048576));
     $gui{spinbuttonCreateHDSize}->set_value(8192.00);
-    my $response = $gui{dialogCreateHD}->run;
-    $gui{dialogCreateHD}->hide;
-    my $diskname = $gui{entryCreateHDName}->get_text();
-    $diskname =~ s/\.vdi$//i;
-    $diskname =~ s/\.vmdk$//i;
-    $diskname =~ s/\.vhd$//i;
-    $diskname =~ s/\.hdd$//i;
-    $diskname =~ s/\.parallels$//i;
 
-    if ($diskname and ($response eq 'ok')) {
-        my %newhd = (diskname   => $diskname,
-                     size       => $gui{spinbuttonCreateHDSize}->get_value_as_int() * 1048576,
-                     allocation => $EMedVar{Standard}, # Standard == Dynamic Allocation
-                     imgformat  => &getsel_combo($gui{comboboxCreateHDFormat}, 1),
-                     location   => &get_path_only(IMachine_getSettingsFilePath($vmc{IMachine})));
+    do {
+        my $response = $gui{dialogCreateHD}->run;
 
-        if ($gui{radiobuttonCreateHDFixed}->get_active()) { $newhd{allocation} = $EMedVar{Fixed}; }
-        elsif ($gui{radiobuttonCreateHDSplit}->get_active()) { $newhd{allocation} = $EMedVar{VmdkSplit2G}; }
-        my $IMedium = &create_new_hd(\%newhd, $gui{dialogEdit});
-        &storage_attach_hd(undef, $IMedium);
-        IMachine_saveSettings($vmc{IMachine});
-        &addrow_log("Settings implicitly saved for $vmc{Name} due to storage attachment.");
-    }
+        if ($response eq 'ok') {
+            # No validation needed on other entries
+            if (!$gui{entryCreateHDName}->get_text()) { &show_err_msg('invalidname', '(Disk Name)'); }
+            else {
+                $gui{dialogCreateHD}->hide;
+                my ($vol, $dir, undef) = &rsplitpath(IMachine_getSettingsFilePath($vmc{IMachine}));
+
+                my %newhd = (diskname   => $gui{entryCreateHDName}->get_text(),
+                             size       => $gui{spinbuttonCreateHDSize}->get_value_as_int() * 1048576,
+                             allocation => 'Standard', # Standard == Dynamic Allocation
+                             imgformat  => &getsel_combo($gui{comboboxCreateHDFormat}, 1),
+                             location   => $vol . $dir);
+
+                if ($gui{radiobuttonCreateHDFixed}->get_active()) { $newhd{allocation} = 'Fixed'; }
+                elsif ($gui{radiobuttonCreateHDSplit}->get_active()) { $newhd{allocation} = 'VmdkSplit2G'; }
+                my $IMedium = &create_new_hd(\%newhd, $gui{dialogEdit});
+
+                if ($IMedium) {
+                    &storage_attach_hd(undef, $IMedium);
+                    IMachine_saveSettings($vmc{IMachine});
+                    &addrow_log("Settings implicitly saved for $vmc{Name} due to storage attachment.");
+                }
+            }
+        }
+        else { $gui{dialogCreateHD}->hide; }
+
+    } until (!$gui{dialogCreateHD}->visible());
 }
 
 # Handle the radio button sensitivity when selecting an image format when
 # creating a hard disk
-sub sens_createhdformat() {
+sub sens_createhdformat {
     my $format = &getsel_combo($gui{comboboxCreateHDFormat}, 1);
     $gui{radiobuttonCreateHDDynamic}->set_active(1);
 
@@ -162,30 +172,39 @@ sub sens_createhdformat() {
 }
 
 # Adds a storage controller to the guest
-sub storage_ctr_add() {
+sub storage_ctr_add {
     my ($widget) = @_;
     my $bus = 'Floppy'; # Assume floppy unless set otherwise
+    my $ctrname = '';
 
     if ($widget eq $gui{menuitemCtrAddIDE}) { $bus = 'IDE'; }
     elsif ($widget eq $gui{menuitemCtrAddSCSI}) { $bus = 'SCSI'; }
     elsif ($widget eq $gui{menuitemCtrAddSATA}) { $bus = 'SATA'; }
     elsif ($widget eq $gui{menuitemCtrAddSAS}) { $bus = 'SAS'; }
+    elsif ($widget eq $gui{menuitemCtrAddUSB}) { $bus = 'USB'; }
+    elsif ($widget eq $gui{menuitemCtrAddNVMe}) { $bus = 'PCIe'; }
 
     my @Controllers = IMachine_getStorageControllers($vmc{IMachine});
     my $exists = 0;
 
+    if ($bus eq 'PCIe') { $ctrname = 'NVMe'; }
+    else { $ctrname = $bus; }
+
     foreach my $ctr (@Controllers) { $exists = 1 if (IStorageController_getBus($ctr) eq $bus) }
 
     if (!$exists) {
-        my $IStorageController = IMachine_addStorageController($vmc{IMachine}, "$bus Controller", $bus);
+        my $IStorageController = IMachine_addStorageController($vmc{IMachine}, $ctrname, $bus);
+        # This works around the controller type defaults defaulting to USB. VB itself doesn't seem to use it.
+        if ($bus eq 'SAS') { IStorageController_setControllerType($IStorageController, 'LsiLogicSas', 0); }
+
         IStorageController_setPortCount($IStorageController, IStorageController_getMinPortCount($IStorageController)); # Controllers have all ports on by default. Set to the minimum
         &fill_list_editstorage($vmc{IMachine});
     }
-    elsif ($exists) { &show_err_msg('ctrallocated', $gui{messagedialogError}); }
+    elsif ($exists) { &show_err_msg('ctrallocated'); }
 }
 
 # Attaches a hard disk to a controller
-sub storage_attach_hd() {
+sub storage_attach_hd {
     my ($widget, $IMedium) = @_;
     my $storref = &getsel_list_editstorage();
     my $attached = 0;
@@ -197,7 +216,7 @@ sub storage_attach_hd() {
     }
 
     # If medium wasn't attached, controller must be full
-    if (!$attached) { &show_err_msg('ctrfull', $gui{messagedialogError}); }
+    if (!$attached) { &show_err_msg('ctrfull'); }
 
     IMachine_saveSettings($vmc{IMachine});
     &addrow_log("Settings implicitly saved for $vmc{Name} due to storage attachment.");
@@ -205,7 +224,7 @@ sub storage_attach_hd() {
 }
 
 # Attaches a DVD to a controller or DVD image to DVD ROM
-sub storage_attach_dvd() {
+sub storage_attach_dvd {
     my ($widget, $IMedium) = @_;
     my $storref = &getsel_list_editstorage();
     my $attached = 0;
@@ -221,7 +240,7 @@ sub storage_attach_dvd() {
         }
 
         # If medium wasn't attached, controller must be full
-        if (!$attached) { &show_err_msg('ctrfull', $gui{messagedialogError}); }
+        if (!$attached) { &show_err_msg('ctrfull'); }
     }
 
     IMachine_saveSettings($vmc{IMachine});
@@ -230,7 +249,7 @@ sub storage_attach_dvd() {
 }
 
 # Attaches a floppy to a controller
-sub storage_attach_floppy() {
+sub storage_attach_floppy {
     my ($widget, $IMedium) = @_;
     my $storref = &getsel_list_editstorage();
     my $attached = 0;
@@ -245,7 +264,7 @@ sub storage_attach_floppy() {
         }
 
         # If medium wasn't attached, controller must be full
-        if (!$attached) { &show_err_msg('ctrfull', $gui{messagedialogError}); }
+        if (!$attached) { &show_err_msg('ctrfull'); }
     }
 
     IMachine_saveSettings($vmc{IMachine});
@@ -254,7 +273,7 @@ sub storage_attach_floppy() {
 }
 
 # Detaches a storage unit or ejects removable media
-sub storage_attach_rem() {
+sub storage_attach_rem {
     my $storref = &getsel_list_editstorage();
 
     if ((($$storref{MediumType} eq 'Floppy') or ($$storref{MediumType} eq 'DVD')) and $$storref{IMedium}) {
@@ -266,61 +285,73 @@ sub storage_attach_rem() {
 }
 
 # Removes a controller from the guest
-sub storage_ctr_rem() {
+sub storage_ctr_rem {
     my $storref = &getsel_list_editstorage();
     my @IMediumAttachments = IMachine_getMediumAttachmentsOfController($vmc{IMachine}, $$storref{ControllerName});
 
-    if (@IMediumAttachments) { &show_err_msg('ctrinuse', $gui{messagedialogWarning}); }
+    if (@IMediumAttachments) { &show_err_msg('ctrinuse'); }
     else { IMachine_removeStorageController($vmc{IMachine}, $$storref{ControllerName}); }
 
     &fill_list_editstorage($vmc{IMachine});
 }
 
 # Sets the controller variant type
-sub storage_ctrtype() {
+sub storage_ctrtype {
     my $storref = &getsel_list_editstorage();
     IStorageController_setControllerType($$storref{IStorageController}, &getsel_combo($gui{comboboxEditStorCtrType}, 0));
 }
 
 # Set / Clear the host I/O cache for the controller
-sub storage_ctrcache() {
+sub storage_ctrcache {
     my $storref = &getsel_list_editstorage();
     IStorageController_setUseHostIOCache($$storref{IStorageController}, $gui{checkbuttonEditStorCache}->get_active());
 }
 
-sub show_attach_menu() {
+sub show_attach_menu {
     my ($widget, $event) = @_;
     $gui{menuAttachAdd}->popup(undef, undef, undef, undef, 0, $event->time) if ($event->button == 1);
     return 0;
 }
 
-sub show_ctr_menu() {
+sub show_ctr_menu {
     my ($widget, $event) = @_; #$event->time
     $gui{menuCtrAdd}->popup(undef, undef, undef, undef, 0, $event->time) if ($event->button == 1);
     return 0;
 }
 
 # Sets a specific port count if manually set by the user
-sub storage_port_count() {
+sub storage_port_count {
     my $storref = &getsel_list_editstorage();
     my $pc = int($gui{adjustmentEditStorPortCount}->get_value());
     IStorageController_setPortCount($$storref{IStorageController}, $pc);
 }
 
 # Flags a hard disk image as being SSD
-sub storage_ssd() {
+sub storage_ssd {
     my $storref = &getsel_list_editstorage();
     IMachine_nonRotationalDevice($vmc{IMachine}, $$storref{ControllerName}, $$storref{Port}, $$storref{Device}, $gui{checkbuttonEditStorSSD}->get_active());
 }
 
 # Flags whether a CD/DVD is allowed to be temporary ejected (ie Live CD)
-sub storage_livecd() {
+sub storage_livecd {
     my $storref = &getsel_list_editstorage();
     IMachine_temporaryEjectDevice($vmc{IMachine}, $$storref{ControllerName}, $$storref{Port}, $$storref{Device}, $gui{checkbuttonEditStorLive}->get_active());
 }
 
+# Flags whether a HD or Optical Device is hot pluggable
+sub storage_hot_pluggable {
+    my $storref = &getsel_list_editstorage();
+    IMachine_setHotPluggableForDevice($vmc{IMachine}, $$storref{ControllerName}, $$storref{Port}, $$storref{Device}, $gui{checkbuttonEditStorHotPluggable}->get_active());
+}
+
+# Flags whether the selected controller is bootable or not
+sub storage_controller_bootable {
+    my $storref = &getsel_list_editstorage();
+    IMachine_setStorageControllerBootable($vmc{IMachine}, $$storref{ControllerName}, $gui{checkbuttonEditStorControllerBootable}->get_active());
+}
+
 # Moves an attachment to a new port
-sub storage_connected_port() {
+sub storage_connected_port {
     my $newdevice = &getsel_combo($gui{comboboxEditStorDevPort}, 1);
     my $newport = &getsel_combo($gui{comboboxEditStorDevPort}, 2);
     my $storref = &getsel_list_editstorage();
@@ -331,17 +362,18 @@ sub storage_connected_port() {
         foreach my $attach (@IMediumAttachment) {
             if ($$attach{port} == $newport and $$attach{device} == $newdevice) {
                 $free = 0;
-                &show_err_msg('existattach', $gui{messagedialogWarning});
+                &show_err_msg('existattach');
                 last;
             }
         }
 
         if ($free) {
-            # At the moment, only SATA support creating actual ports
-            if (IStorageController_getPortCount($$storref{IStorageController}) < ($newport + 1) and $$storref{Bus} eq 'SATA') {
+            # SATA, SAS, NVMe support changing the ports
+            if ($$storref{Bus} eq 'SATA' or $$storref{Bus} eq 'SAS' or $$storref{Bus} eq 'PCIe') {
+            if (IStorageController_getPortCount($$storref{IStorageController}) < ($newport + 1)) {
                 IStorageController_setPortCount($$storref{IStorageController}, $newport + 1) ;
             }
-
+            }
             IMachine_detachDevice($vmc{IMachine}, $$storref{ControllerName}, $$storref{Port}, $$storref{Device});
             IMachine_attachDevice($vmc{IMachine}, $$storref{ControllerName}, $newport, $newdevice, $$storref{MediumType}, $$storref{IMedium});
             IMachine_saveSettings($vmc{IMachine});
@@ -351,109 +383,12 @@ sub storage_connected_port() {
     }
 }
 
-# Handles the sensitive if a SATA controller is selected
-sub sens_editstor_sata_ctr() {
-    $gui{labelEditStorPortCount}->show();
-    $gui{spinbuttonEditStorPortCount}->show();
-    $gui{menuitemAttachHD}->set_sensitive(1);
-    $gui{menuitemAttachDVD}->set_sensitive(1);
-    $gui{menuitemAttachFloppy}->set_sensitive(0);
-}
-
-sub sens_editstor_ide_ctr() {
-    $gui{menuitemAttachHD}->set_sensitive(1);
-    $gui{menuitemAttachDVD}->set_sensitive(1);
-    $gui{menuitemAttachFloppy}->set_sensitive(0);
-}
-
-sub sens_editstor_sas_ctr() {
-    $gui{menuitemAttachHD}->set_sensitive(1);
-    $gui{menuitemAttachDVD}->set_sensitive(0);
-    $gui{menuitemAttachFloppy}->set_sensitive(0);
-}
-
-sub sens_editstor_scsi_ctr() {
-    $gui{menuitemAttachHD}->set_sensitive(1);
-    $gui{menuitemAttachDVD}->set_sensitive(0);
-    $gui{menuitemAttachFloppy}->set_sensitive(0);
-}
-
-sub sens_editstor_floppy_ctr() {
-    $gui{menuitemAttachHD}->set_sensitive(0);
-    $gui{menuitemAttachDVD}->set_sensitive(0);
-    $gui{menuitemAttachFloppy}->set_sensitive(1);
-}
-
-sub sens_editstor_allctr() {
-    $gui{frameEditStorAttr}->show();
-    $gui{buttonEditStorAddAttach}->set_sensitive(1);
-    $gui{buttonEditStorRemoveAttach}->set_sensitive(0);
-    $gui{buttonEditStorRemoveCtr}->set_sensitive(1);
-    $gui{labelEditStorCtrName}->show();
-    $gui{entryEditStorCtrName}->show();
-    $gui{labelEditStorCtrType}->show();
-    $gui{comboboxEditStorCtrType}->show();
-    $gui{checkbuttonEditStorCache}->show();
-    $gui{labelEditStorDevPort}->hide();
-    $gui{comboboxEditStorDevPort}->hide();
-    $gui{checkbuttonEditStorSSD}->hide();
-    $gui{checkbuttonEditStorLive}->hide();
-    $gui{labelEditStorPortCount}->hide();
-    $gui{spinbuttonEditStorPortCount}->hide();
-}
-
-# Handle widget sensitivity specify to HD media selection
-sub sens_editstor_hd() {
-    $gui{buttonEditStorAddAttach}->set_sensitive(0);
-    $gui{checkbuttonEditStorSSD}->show();
-    $gui{checkbuttonEditStorLive}->hide();
-    $gui{menuitemAttachHD}->set_sensitive(1);
-    $gui{menuitemAttachDVD}->set_sensitive(0);
-    $gui{menuitemAttachFloppy}->set_sensitive(0);
-}
-
-# Handle widget sensitivity specify to DVD media selection
-sub sens_editstor_dvd() {
-    $gui{buttonEditStorAddAttach}->set_sensitive(1);
-    $gui{checkbuttonEditStorSSD}->hide();
-    $gui{checkbuttonEditStorLive}->show();
-    $gui{menuitemAttachHD}->set_sensitive(0);
-    $gui{menuitemAttachDVD}->set_sensitive(1);
-    $gui{menuitemAttachFloppy}->set_sensitive(0);
-}
-
-# Handle widget sensitivity specify to floppy media selection
-sub sens_editstor_floppy() {
-    $gui{buttonEditStorAddAttach}->set_sensitive(1);
-    $gui{checkbuttonEditStorSSD}->hide();
-    $gui{checkbuttonEditStorLive}->hide();
-    $gui{menuitemAttachHD}->set_sensitive(0);
-    $gui{menuitemAttachDVD}->set_sensitive(0);
-    $gui{menuitemAttachFloppy}->set_sensitive(1);
-}
-
-# Handle widget sensitivity common to all media selection
-sub sens_editstor_allmedia() {
-    $gui{frameEditStorAttr}->show();
-    $gui{buttonEditStorRemoveAttach}->set_sensitive(1);
-    $gui{buttonEditStorRemoveCtr}->set_sensitive(0);
-    $gui{labelEditStorCtrName}->hide();
-    $gui{entryEditStorCtrName}->hide();
-    $gui{labelEditStorCtrType}->hide();
-    $gui{comboboxEditStorCtrType}->hide();
-    $gui{checkbuttonEditStorCache}->hide();
-    $gui{labelEditStorDevPort}->show();
-    $gui{comboboxEditStorDevPort}->show();
-    $gui{labelEditStorPortCount}->hide();
-    $gui{spinbuttonEditStorPortCount}->hide();
-}
-
 # Sets the sensitive of widgets if no storage item is selected
-sub sens_editstor_nosel() {
+sub sens_editstor_nosel {
     $gui{frameEditStorAttr}->hide(); # Hides all the attributed widgets in one go
     $gui{buttonEditStorAddAttach}->set_sensitive(0);
     $gui{buttonEditStorRemoveAttach}->set_sensitive(0);
     $gui{buttonEditStorRemoveCtr}->set_sensitive(0);
 }
 
-1;
\ No newline at end of file
+1;
